package com.gct.dataengine.register.service.impl;

import com.alibaba.fastjson.JSONObject;
import com.gct.dataengine.register.entity.Collector;
import com.gct.dataengine.register.entity.CollectorTaskInfoVo;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Service;

import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantLock;
import java.util.stream.Collectors;

/**
 * @author xusirui 2022/4/8
 */
@Service
@Slf4j
public class CollectorTaskManage {

    @Autowired
    RedisTemplate<String, Object> redisTemplate;

    public static final String collectorSetName = "collectorSet";

    public static final String taskCollectorMapName = "taskCollectorMap";

    public static final String taskVersionMapName = "taskVersionMap";

    public static final String nullStr = "null";

    public static final String versionName = "version";

    private ReentrantLock lock = new ReentrantLock();

    /**
     * @param collector
     */
    public Boolean heartBeat(Collector collector) {
        try{
            redisTemplate.opsForZSet().add(collectorSetName, collector, collector.getTimeOut()*1000 + System.currentTimeMillis());
        }catch (Exception e){
            e.printStackTrace();
            return false;
        }
        return true;
    }

    /**
     * 1 minute 5s
     * @return device list
     */
    public void unbind(){
        //unbind
        long nowTimePoint = System.currentTimeMillis();
        Set<Object> outlineCollectorIds = redisTemplate.opsForZSet()
                .rangeByScore(collectorSetName, 0, nowTimePoint)
                .stream()
                .map(x->JSONObject.parseObject(x.toString(),Collector.class).getId())
                .collect(Collectors.toSet());
        try{
            redisTemplate.multi();
            redisTemplate.opsForHash().entries(taskCollectorMapName).forEach((taskId,CollectorId)->{
                if (outlineCollectorIds.contains(CollectorId)){
                    redisTemplate.opsForHash().put(collectorSetName,taskId,nullStr);
                }
            });
            //clear outline collector
            redisTemplate.opsForZSet().removeRangeByScore(collectorSetName,0,nowTimePoint);
            redisTemplate.exec();
            if (!outlineCollectorIds.isEmpty()) log.info("Collector list has clear outline's collector {}",outlineCollectorIds);
        }catch (Exception e){
            redisTemplate.discard();
        }
    }

    public List<Collector> getOnlineCollectors() {
        return redisTemplate.opsForZSet().rangeByScore(collectorSetName,System.currentTimeMillis(), Long.MAX_VALUE)
                .stream()
                .map(x-> JSONObject.parseObject(x.toString(),Collector.class))
                .collect(Collectors.toList());
    }

    public Map<Object, Object> getCollectorTaskMap(List<Collector> onlineCollectors){
        Set<String> onlineCollectorIds = onlineCollectors.stream().map(Collector::getId).collect(Collectors.toSet());
        Map<Object, Object> map = redisTemplate.opsForHash().entries(taskCollectorMapName);
        map.entrySet().stream()
                .filter(entry->!onlineCollectorIds.contains(entry.getValue()))
                .collect(Collectors.toList())
                .forEach(entry->map.put(entry.getKey(),nullStr));
        return map;
    }

    /**
     * push or update
     * @param taskId
     */
    public void pushTask(String taskId,String version){
        Object originalVersion = redisTemplate.opsForHash().get(taskVersionMapName, taskId);
        if (!version.equals(originalVersion)){
            redisTemplate.opsForHash().put(taskCollectorMapName,taskId,nullStr);
            redisTemplate.opsForHash().put(taskVersionMapName,taskId,version);
        }
    }

    public void removeTask(String taskId) {
        redisTemplate.opsForHash().delete(taskCollectorMapName,taskId);
        redisTemplate.opsForHash().delete(taskVersionMapName,taskId);
    }

    public Set<Object> getAllTask(){
        return redisTemplate.opsForHash().entries(taskCollectorMapName).keySet();
    }

    /**
     * @param
     * @return collector list and collector task map
     */
    public CollectorTaskInfoVo pullCollectorTaskInfo(){
        //client balance needn't unbind before get map and list
        //no new version,old data
        String version = redisTemplate.opsForValue().get(versionName).toString();
        List<Collector> onlineCollectors = getOnlineCollectors();
        Map<Object,Object> collectorTaskMap = getCollectorTaskMap(onlineCollectors);
        return new CollectorTaskInfoVo(onlineCollectors,collectorTaskMap,version);
    }

    @SneakyThrows
    public Boolean bind(String collectorId, String taskId, String version){
        if (!lock.tryLock(10, TimeUnit.MILLISECONDS))return false;
        boolean res = true;
        if (version.equals(redisTemplate.opsForValue().get(versionName))){
            try{
                redisTemplate.watch(versionName);
                redisTemplate.multi();
                redisTemplate.opsForValue().set(versionName,UUID.randomUUID());
                redisTemplate.opsForHash().put(taskCollectorMapName,taskId,collectorId);
                List<Object> exec = redisTemplate.exec();
                if (exec.isEmpty())return false;
            }catch (Exception e){
                e.printStackTrace();
                redisTemplate.discard();
            }finally {
                redisTemplate.unwatch();
            }

        }else {
            res = false;
        }
        lock.unlock();
        return res;
    }

}
